from __future__ import annotations

from typing import Any, Iterator, Optional

import numpy as np

from ..exceptions import ValidationError
from ..types import FloatArray, PType
from .boxes import Boxes
from .single_box import SingleBox


def _is_2d(x: FloatArray) -> None:
    if x.ndim != 2:
        raise ValidationError("A frame boxes object must be a 2D array")


class FrameBoxes(Boxes):
    """
    2D dense data arrays for boxes.

    It inherits from :class:`.Boxes`.
    The underlying NumPy array should have shape ``(n_boxes, 4)``.

    Parameters
    ----------
    data : FloatArray
        A numpy array of dense floats
    ptype : PType
        The parameterization of the dense data
    image_width : int, optional
        The width of the relevant image
    image_height : int, optional
        The height of the relevant image

    Example
    -------
    >>> import numpy as np
    >>> from sparrow_datums import FrameBoxes, PType
    >>> boxes = FrameBoxes(np.array([[0, 0, 1, 1]]))
    >>> for box in boxes: print(box)
    [0 0 1 1]
    """

    empty_shape: tuple[int, int] = (0, 4)

    def validate(self) -> None:
        """Check validity of boxes array."""
        super().validate()
        _is_2d(self)

    def __iter__(self) -> Iterator[SingleBox]:
        """Yield SingoeBox objects for each box."""
        for box in self.view(Boxes):
            yield box.view(SingleBox)

    @classmethod
    def from_single_box(cls: type["FrameBoxes"], box: SingleBox) -> "FrameBoxes":
        """Create a FrameBoxes object from a SingleBox."""
        return cls(
            box.array[None, :],
            ptype=box.ptype,
            **box.metadata_kwargs,
        )

    @classmethod
    def from_single_boxes(
        cls: type["FrameBoxes"],
        boxes: list[SingleBox],
        ptype: PType = PType.unknown,
        image_width: Optional[int] = None,
        image_height: Optional[int] = None,
        **kwargs: dict[str, Any],
    ) -> "FrameBoxes":
        """Create a FrameBoxes object from a list of SingleBox objects."""
        return cls(
            np.stack([box.array for box in boxes]),
            ptype=ptype,
            image_width=image_width,
            image_height=image_height,
        )

    def add_box(self, box: SingleBox) -> "FrameBoxes":
        """Concatenate a single box."""
        if self.ptype != box.ptype:
            raise ValidationError(
                "SingleBox with different PType cannot be concatenated"
            )
        if self.metadata_kwargs != box.metadata_kwargs:
            raise ValidationError(
                "SingleBox with different metadata cannot be concatenated"
            )
        return FrameBoxes(
            np.concatenate([self.array, box.array[None]]),
            ptype=self.ptype,
            **self.metadata_kwargs,
        )

    def get_single_box(self, i: int) -> SingleBox:
        """Get the ith element as a SingleBox."""
        result: SingleBox = self.view(Boxes)[i].view(SingleBox)
        return result



from ..exceptions import ValidationError
from ..types import FloatArray
from .boxes import Boxes


def _is_1d(x: FloatArray) -> None:
    if x.ndim != 1:
        raise ValidationError("Single box must be a 1D array")


class SingleBox(Boxes):
    """
    1D dense data arrays for a box.

    It inherits from :class:`.Boxes`.
    The underlying NumPy array should have shape ``(4,)``.

    Parameters
    ----------
    data : FloatArray
        A numpy array of dense floats
    ptype : PType
        The parameterization of the dense data
    image_width : int, optional
        The width of the relevant image
    image_height : int, optional
        The height of the relevant image
    """

    def validate(self) -> None:
        """Check validity of boxes array."""
        super().validate()
        _is_1d(self)



from __future__ import annotations

import json
from operator import itemgetter
from pathlib import Path
from typing import Any, Iterator, Optional, Union

import numpy as np
import numpy.typing as npt

from ..exceptions import ValidationError
from ..types import PType
from .augmented_boxes import AugmentedBoxes
from .frame_boxes import FrameBoxes, _is_2d
from .single_augmented_box import SingleAugmentedBox


class FrameAugmentedBoxes(AugmentedBoxes):
    """
    2D dense data arrays for augmented boxes.

    The data contain ``[boxes, scores, labels]`` components.
    It inherits from :class:`.AugmentedBoxes`.
    The underlying NumPy array should have shape ``(n_boxes, 6)``,
    with 4 dimensions reserved for boxes, and the last two for
    scores and labels respectively.

    Parameters
    ----------
    data : FloatArray
        A numpy array of dense floats
    ptype : PType
        The parameterization of the dense data
    image_width : int, optional
        The width of the relevant image
    image_height : int, optional
        The height of the relevant image

    Example
    -------
    >>> import numpy as np
    >>> from sparrow_datums import FrameAugmentedBoxes, PType
    >>> augmented_boxes = FrameAugmentedBoxes(np.array([[0, 0, 1, 1, 0.5, 3]]))
    >>> for box in augmented_boxes: print(box)
    [0.  0.  1.  1.  0.5 3. ]
    """

    empty_shape: tuple[int, int] = (0, 6)

    def validate(self) -> None:
        """Check validity of boxes array."""
        super().validate()
        _is_2d(self)

    def __iter__(self) -> Iterator[SingleAugmentedBox]:
        """Yield SingleAugmentedBox objects for each box."""
        for box in self.view(AugmentedBoxes):
            yield box.view(SingleAugmentedBox)

    def to_darwin_dict(
        self,
        filename: str,
        path: str = "/",
        label_names: Optional[list[str]] = None,
    ) -> dict[str, Any]:
        """Serialize boxes to a Darwin annotation dict."""
        if label_names is None:
            label_names = ["Unknown"] * (self.labels.max() + 1)
        return {
            "image": {"filename": filename, "path": path},
            "annotations": [
                {
                    "bounding_box": {
                        "x": float(box.x),
                        "y": float(box.y),
                        "w": float(box.w),
                        "h": float(box.h),
                    },
                    "name": box.names(label_names),
                }
                for box in self.to_absolute()
            ],
        }

    def to_darwin_file(
        self,
        output_path: Union[str, Path],
        filename: str,
        path: str = "/",
        label_names: Optional[list[str]] = None,
    ) -> None:
        """
        Write Darwin annotation dict to disk.

        Parameters
        ----------
        output_path : str
            The full path (including filename) where the Darwin annotation will be written on disk
        filename : str
            The fimename of the image or video as it's referenced within the Darwin
        path : str, optional
            The path of the file within Darwin. The path starts at root ("/"),
            and expands including the folders in which the resource is stored within Darwin.
        label_names : list of str, optional
            The names of the labeled classes annotated in the image or video. If None,
            "Unknown" will be used.
        """
        with open(output_path, "w") as f:
            f.write(
                json.dumps(
                    self.to_darwin_dict(filename, path=path, label_names=label_names)
                )
            )

    @classmethod
    def from_darwin_dict(
        cls: type["FrameAugmentedBoxes"],
        darwin_dict: dict[str, Any],
        label_names: list[str] = [],
    ) -> "FrameAugmentedBoxes":
        """Create FrameAugmentedBoxes from a serialized Darwin dict."""
        label_names_map = {name: float(idx) for idx, name in enumerate(label_names)}
        image_width, image_height = itemgetter("width", "height")(darwin_dict["image"])
        boxes = []
        score = 1.0
        for annotation in darwin_dict["annotations"]:
            if "bounding_box" not in annotation:
                continue
            x, y, w, h = itemgetter("x", "y", "w", "h")(annotation["bounding_box"])
            label = label_names_map.get(annotation["name"], -1.0)
            boxes.append([x, y, w, h, score, label])
        data: npt.NDArray[np.float64]
        if len(boxes):
            data = np.array(boxes).astype("float64")
        else:
            data = np.zeros((0, 6), "float64")
        return cls(
            data,
            ptype=PType.absolute_tlwh,
            image_width=image_width,
            image_height=image_height,
        )

    @classmethod
    def from_darwin_file(
        cls: type["FrameAugmentedBoxes"],
        path: Union[str, Path],
        label_names: list[str] = [],
    ) -> "FrameAugmentedBoxes":
        """Read FrameAugmentedBoxes from Darwin dict on disk."""
        with open(path) as f:
            darwin_dict = json.loads(f.read())
        return cls.from_darwin_dict(darwin_dict, label_names=label_names)

    @classmethod
    def from_dict(
        cls: type["FrameAugmentedBoxes"],
        chunk_dict: dict[str, Any],
        dims: Optional[int] = None,
    ) -> "FrameAugmentedBoxes":
        """Create chunk from chunk dict."""
        return super().from_dict(chunk_dict, dims=6)

    @classmethod
    def from_single_box(
        cls: type["FrameAugmentedBoxes"], box: SingleAugmentedBox
    ) -> "FrameAugmentedBoxes":
        """Create a FrameBoxes object from a SingleBox."""
        return cls(
            box.array[None, :],
            ptype=box.ptype,
            **box.metadata_kwargs,
        )

    def add_box(self, box: SingleAugmentedBox) -> "FrameAugmentedBoxes":
        """Concatenate a single augmented box."""
        if self.ptype != box.ptype:
            raise ValidationError(
                "SingleAugmentedBox with different PType cannot be concatenated"
            )
        if self.metadata_kwargs != box.metadata_kwargs:
            raise ValidationError(
                "SingleAugmentedBox with different metadata cannot be concatenated"
            )
        return FrameAugmentedBoxes(
            np.concatenate([self.array, box.array[None]]),
            ptype=self.ptype,
            **self.metadata_kwargs,
        )

    def get_single_box(self, i: int) -> SingleAugmentedBox:
        """Get the ith element as a SingleAugmentedBox."""
        result: SingleAugmentedBox = self.view(AugmentedBoxes)[i].view(
            SingleAugmentedBox
        )
        return result

    def to_frame_boxes(self) -> FrameBoxes:
        """Drop augmented part of the data."""
        return FrameBoxes(
            self.array[..., :4],
            ptype=self.ptype,
            **self.metadata_kwargs,
        )



from __future__ import annotations

from pathlib import Path
from typing import Iterator

import numpy as np

from ..chunk import T
from ..exceptions import ValidationError
from ..tracking import Tracking
from ..types import PType
from .types import ChunkPath, Footer, Header


class ChunkStreamReader:
    """A class for reading chunk streams from disk."""

    def __init__(self, manifest_path: str | Path, chunk_type: type[T]) -> None:
        self.manifest_path = Path(manifest_path)
        if not issubclass(chunk_type, Tracking):
            raise ValidationError(f"Invalid stream chunk type {chunk_type}")
        self.chunk_type = chunk_type
        header, chunk_paths, footer = self.parse_manifest()
        self.header = header
        self.chunk_paths = chunk_paths
        self.footer = footer

    def parse_manifest(self) -> tuple[Header, list[ChunkPath], Footer]:
        """Write the manifest with all the chunks."""
        with open(self.manifest_path) as f:
            value_strings = f.read().split("\n")
        if len(value_strings) < 2:
            raise ValueError(f"Invalid manifest at {self.manifest_path}")
        header = Header.from_json(value_strings[0])
        footer = Footer.from_json(value_strings[-1])
        chunk_paths = list(map(ChunkPath.from_json, value_strings[1:-1]))
        return header, chunk_paths, footer

    def __len__(self) -> int:
        """Get length of chunk stream."""
        return len(self.chunk_paths)

    def __getitem__(self, idx: int) -> T:
        """Get a chunk from a stream."""
        chunk_path_object = self.chunk_paths[idx]
        chunk_path = self.manifest_path.parent / chunk_path_object.path
        try:
            return self.chunk_type.from_file(chunk_path)
        except ValidationError:
            chunk = self.chunk_type.empty(
                ptype=PType[self.header.ptype],
                image_width=self.header.image_width,
                image_height=self.header.image_height,
                fps=self.header.fps,
                start_time=chunk_path_object.start_time,
            )
            new_shape = list(chunk.shape)
            new_shape[0] = round(chunk_path_object.duration * self.header.fps)
            return chunk.pad(tuple(new_shape))

    def __iter__(self) -> Iterator[T]:
        """Yield chunks from a stream."""
        for i in range(len(self)):
            yield self[i]

    def concat(self) -> T:
        n_frames = 0
        object_ids = set()
        n_objects = 0
        for chunk in self:
            n_frames += len(chunk)
            n_objects = max(n_objects, chunk.shape[1])
            object_ids |= set(chunk.object_ids)
        object_ids = sorted(object_ids)
        n_objects = max(n_objects, len(object_ids))
        data = np.zeros((n_frames, n_objects, chunk.shape[-1])) * np.nan
        frame_idx = 0
        for chunk in self:
            object_indices = np.array(
                [object_ids.index(object_id) for object_id in chunk.object_ids]
            )
            if len(object_indices) < chunk.shape[1]:
                object_indices = np.arange(chunk.shape[1])
            data[frame_idx : frame_idx + len(chunk), object_indices] = chunk.data
            frame_idx += len(chunk)
        return self.chunk_type(
            data,
            ptype=chunk.ptype,
            **chunk.metadata_kwargs,
        )



import os
import tempfile

import numpy as np

from sparrow_datums import BoxTracking

from .reader import ChunkStreamReader
from .writer import ChunkStreamWriter


def test_chunk_stream_reader():
    with tempfile.TemporaryDirectory() as dir:
        manifest_path = os.path.join(dir, "stream.jsonl")
        writer = ChunkStreamWriter(manifest_path, BoxTracking, fps=1)
        a = BoxTracking(np.ones((2, 2, 4)), fps=1, start_time=0)
        b = BoxTracking(np.ones((2, 2, 4)), fps=1, start_time=a.duration)
        c = BoxTracking(np.ones((2, 1, 4)), fps=1, start_time=a.duration + b.duration)
        writer.add_chunk(a)
        writer.add_chunk(b)
        writer.add_chunk(c)
        writer.close()
        reader = ChunkStreamReader(manifest_path, BoxTracking)
        assert reader.header.fps == 1
        assert len(reader) == 3
        assert isinstance(reader[1], BoxTracking)
        assert all(isinstance(chunk, BoxTracking) for chunk in reader)
        assert reader.concat().shape == (6, 2, 4)


def test_chunk_stream_reader_concat_with_object_ids():
    with tempfile.TemporaryDirectory() as dir:
        manifest_path = os.path.join(dir, "stream.jsonl")
        writer = ChunkStreamWriter(manifest_path, BoxTracking, fps=1)
        a = BoxTracking(np.ones((2, 2, 4)), fps=1, start_time=0, object_ids=["1", "2"])
        b = BoxTracking(
            np.ones((2, 2, 4)), fps=1, start_time=a.duration, object_ids=["2", "3"]
        )
        writer.add_chunk(a)
        writer.add_chunk(b)
        writer.close()
        reader = ChunkStreamReader(manifest_path, BoxTracking)
        assert reader.concat().shape == (4, 3, 4)



import numpy as np
import pytest

from ..exceptions import ValidationError
from .box_tracking import BoxTracking


def test_box_tracking_chunk_requires_3_dimensions():
    with pytest.raises(ValidationError):
        BoxTracking(np.ones((2, 4)))
    with pytest.raises(ValidationError):
        BoxTracking(np.ones((2, 2, 2)))



"""BoxTracking chunk."""
from __future__ import annotations

from typing import Iterator, Optional

import numpy as np

from ..boxes import Boxes, FrameBoxes
from ..types import PType
from .tracking import Tracking


class BoxTracking(Tracking, Boxes):
    """
    Dense data arrays for box tracking.

    It inherits from :class:`.Boxes`.
    The underlying NumPy array should have shape ``(n_frames, n_objects, 4)``.

    Parameters
    ----------
    data : FloatArray
        A numpy array of dense floats
    ptype : PType
        The parameterization of the dense data
    image_width : int, optional
        The width of the relevant image
    image_height : int, optional
        The height of the relevant image
    fps : float, optional
        The framerate of the chunk data (if tracking)
    object_ids:  list[str], optional
        Identifiers for the objects (if tracking)
    """

    empty_shape: tuple[int, int, int] = (0, 0, 4)

    def validate(self) -> None:
        """Validate tracking shape and boxes."""
        super().validate()
        Boxes.validate(self)

    def __iter__(self) -> Iterator[FrameBoxes]:
        """Yield FrameBoxes objects for each frame."""
        for box in self.view(Boxes):
            yield box.view(FrameBoxes)

    @classmethod
    def from_frame_boxes(
        cls: type["BoxTracking"],
        boxes: list[FrameBoxes],
        ptype: PType = PType.unknown,
        image_width: Optional[int] = None,
        image_height: Optional[int] = None,
        fps: Optional[float] = None,
        object_ids: Optional[list[str]] = None,
    ) -> "BoxTracking":
        """
        Create an BoxTracking chunk from a list of FrameBoxes objects.

        This is typically used for storing detections on a video.
        """
        max_boxes = max(map(len, boxes))
        data = np.zeros((len(boxes), max_boxes, 4)) * np.nan
        for i, frame in enumerate(boxes):
            data[i, : len(frame)] = frame.array
        return cls(
            data,
            ptype=ptype,
            image_width=image_width,
            image_height=image_height,
            fps=fps,
            object_ids=object_ids,
        )



import numpy as np
import pytest

from ..exceptions import ValidationError
from ..types import PType
from .single_augmented_box import SingleAugmentedBox


def test_single_augmented_box_conversion_creates_single_augmented_box():
    box_a = SingleAugmentedBox(np.ones(6), PType.absolute_tlwh)
    box_b = box_a.to_tlbr()
    assert box_b.is_tlbr
    assert isinstance(box_b, SingleAugmentedBox)


def test_single_augmented_box_with_bad_shape_throws_value_error():
    with pytest.raises(ValidationError):
        SingleAugmentedBox(np.ones((2, 6)))
    with pytest.raises(ValidationError):
        SingleAugmentedBox(np.ones(4))


def test_score_alias_is_float():
    box = SingleAugmentedBox(np.ones(6))
    assert isinstance(box.score, float)
    assert box.score == 1


def test_label_alias_is_int():
    box = SingleAugmentedBox(np.ones(6))
    assert isinstance(box.label, int)
    assert box.label == 1


def test_to_single_box_preserves_as_much_data_as_possible():
    single_augmented_box = SingleAugmentedBox(np.array([0, 0, 1, 1, 0.4, 7]))
    single_box = single_augmented_box.to_single_box()
    np.testing.assert_equal(single_augmented_box.array[:4], single_box.array)
    assert single_augmented_box.ptype == single_box.ptype



from __future__ import annotations

import numpy as np
import numpy.typing as npt

from ..exceptions import ValidationError
from ..types import FloatArray
from .boxes import Boxes


class AugmentedBoxes(Boxes):
    """
    Dense data arrays for augmented boxes.

    The data contain ``[boxes, scores, labels]`` components.
    It inherits from :class:`.Boxes`.
    The underlying NumPy array should have shape ``(..., 6)``,
    with 4 dimensions reserved for boxes, and the last two for
    scores and labels respectively.

    Parameters
    ----------
    data : FloatArray
        A numpy array of dense floats
    ptype : PType
        The parameterization of the dense data
    image_width : int, optional
        The width of the relevant image
    image_height : int, optional
        The height of the relevant image

    Example
    -------
    >>> import numpy as np
    >>> from sparrow_datums import AugmentedBoxes, PType
    >>> augmented_boxes = AugmentedBoxes(np.array([[0, 0, 1, 1, 0.5, 3]]))
    >>> augmented_boxes.scores
    array([0.5])
    >>> augmented_boxes.labels
    array([3])
    """

    def validate(self) -> None:
        """Check validity of boxes array."""
        if not self.shape or self.shape[-1] != 6:
            raise ValidationError("AugmentedBoxes arrays must have size-6 dimensions")
        if not np.all(np.nan_to_num(np.mod(self.array[..., -1], 1)) == 0):
            raise ValidationError("labels must be whole number class indices")
        nonan_scores = np.nan_to_num(self.scores)
        if not np.all(nonan_scores >= 0) or not np.all(nonan_scores <= 1):
            raise ValidationError("scores array must be floats in [0, 1]")

    @property
    def scores(self) -> FloatArray:
        """Confidence scores."""
        result: npt.NDArray[np.float64] = self.array[..., -2]
        return result

    @property
    def labels(self) -> npt.NDArray[np.int64]:
        """Class label indices."""
        result: npt.NDArray[np.int64] = np.nan_to_num(
            self.array[..., -1], nan=-1
        ).astype(np.int64)
        return result

    def names(self, label_names: list[str]) -> npt.NDArray[np.str_]:
        """Map class label indices to string names."""
        result: npt.NDArray[np.str_] = np.array(label_names)[self.labels].astype(
            np.str_
        )
        return result



"""Base class for dense data arrays with metadata."""
from __future__ import annotations

import abc
import gzip
import json
from pathlib import Path
from typing import Any, TypeVar

import numpy as np

from .exceptions import ValidationError
from .types import FloatArray, PType

T = TypeVar("T", bound="Chunk")


class Chunk(FloatArray):
    """
    Base class for dense data arrays with metadata.

    Parameters
    ----------
    data : FloatArray
        A numpy array of dense floats
    ptype : PType
        The parameterization of the dense data
    image_width : int, optional
        The width of the relevant image
    image_height : int, optional
        The height of the relevant image
    fps : float, optional
        The framerate of the chunk data (if tracking)
    object_ids:  list[str], optional
        Identifiers for the objects (if tracking)
    start_time: float, optional
        Start time of chunk with respect to video (if tracking)
    """

    empty_shape: tuple[int, ...]

    def __new__(
        cls: type[T],
        data: FloatArray,
        ptype: PType = PType.unknown,
        image_width: int | None = None,
        image_height: int | None = None,
        fps: float | None = None,
        object_ids: list[str] | None = None,
        start_time: float | None = None,
    ) -> T:
        """Instantiate a new chunk."""
        obj: T = np.asarray(data).view(cls)
        obj.ptype = ptype
        obj._image_width = image_width
        obj._image_height = image_height
        obj._fps = fps
        obj._object_ids = object_ids
        obj._start_time = start_time
        obj._scale = None
        return obj

    def __array_finalize__(self, obj: T | None) -> None:
        """Instantiate a new chunk created from a view."""
        if obj is None:
            return
        self.ptype: PType = getattr(obj, "ptype", PType.unknown)
        self._image_width: float | None = getattr(obj, "_image_width", None)
        self._image_height: float | None = getattr(obj, "_image_height", None)
        self._fps: float | None = getattr(obj, "_fps", None)
        self._object_ids: list[str] | None = getattr(obj, "_object_ids", None)
        self._start_time: float | None = getattr(obj, "_start_time", None)
        self._scale: FloatArray | None = getattr(obj, "_scale", None)
        self.validate()

    @abc.abstractmethod
    def validate(self) -> None:
        """Raise ValidationError for incorrect shape or values."""
        raise NotImplementedError

    @classmethod
    def empty(
        cls: type[T],
        ptype: PType = PType.unknown,
        image_width: int | None = None,
        image_height: int | None = None,
        fps: float | None = None,
        object_ids: list[str] | None = None,
        start_time: float | None = None,
    ) -> T:
        """Create an empty chunk."""
        data = np.zeros(cls.empty_shape) * np.nan
        return cls(
            data,
            ptype=ptype,
            image_width=image_width,
            image_height=image_height,
            fps=fps,
            object_ids=object_ids,
            start_time=start_time,
        )

    @property
    def array(self) -> FloatArray:
        """Dense data as an ndarray."""
        return self.view(np.ndarray)

    @property
    def image_width(self) -> float:
        """Image width."""
        if self._image_width is None:
            raise ValidationError("image_width not set")
        return self._image_width

    @property
    def image_height(self) -> float:
        """Image height."""
        if self._image_height is None:
            raise ValidationError("image_height not set")
        return self._image_height

    @property
    def fps(self) -> float:
        """Frames per second."""
        if self._fps is None:
            raise ValidationError("fps not set")
        return self._fps

    @property
    def object_ids(self) -> list[str]:
        """List of IDs for tracked objects."""
        if self._object_ids is None:
            return []
        return self._object_ids

    @property
    def start_time(self) -> float:
        """Start time of chunk with respect to a video."""
        if self._start_time is None:
            raise ValidationError("start_time not set")
        return self._start_time

    @property
    def scale(self) -> FloatArray:
        """Scaling array."""
        if self._scale is None:
            width = self.image_width
            height = self.image_height
            self._scale = np.array([width, height, width, height])
        return self._scale

    @property
    def duration(self) -> float:
        """Duration of the chunk in seconds."""
        raise NotImplementedError

    @property
    def metadata_kwargs(self) -> dict[str, Any]:
        """Metadata kwargs for downstream constructors."""
        return {
            "image_width": self._image_width,
            "image_height": self._image_height,
            "fps": self._fps,
            "object_ids": self._object_ids,
            "start_time": self._start_time,
        }

    def pad(self, shape: tuple[int, ...]) -> T:
        """Pad a chunk with NaN values."""
        try:
            assert len(shape) == self.ndim
            assert all(a <= b for a, b in zip(self.shape, shape))
        except AssertionError:
            raise ValidationError(
                f"Shape should be compatible with {self.shape} but got {shape}."
            )
        pad_width = tuple((0, b - a) for a, b in zip(self.shape, shape))
        data = np.pad(self.array, pad_width, "constant", constant_values=np.nan)
        return self.__class__(
            data,
            ptype=self.ptype,
            **self.metadata_kwargs,
        )

    def to_dict(self) -> dict[str, Any]:
        """Serialize chunk to a dict."""
        return {
            "data": np.where(np.isnan(self.array), np.array(None), self.array).tolist(),
            "classname": self.__class__.__name__,
            "ptype": self.ptype.name,
            **self.metadata_kwargs,
        }

    def to_file(self, path: str | Path) -> None:
        """Write serialized chunk to disk."""
        if not str(path).endswith(".json.gz"):
            raise ValidationError("Chunk file name must end with .json.gz")
        with gzip.open(path, "wt") as f:
            f.write(json.dumps(self.to_dict()))

    @classmethod
    def from_dict(
        cls: type[T], chunk_dict: dict[str, Any], dims: int | None = None
    ) -> T:
        """Create chunk from chunk dict."""
        data: FloatArray
        if len(chunk_dict["data"]) == 0 and dims:
            if "tracking" in cls.__name__.lower():
                data = np.zeros((0, 0, dims), "float64")
            else:
                data = np.zeros((0, dims), "float64")
        else:
            data = np.array(chunk_dict["data"]).astype("float64")
        data[data == None] = np.nan
        return cls(
            data,
            ptype=PType(chunk_dict["ptype"]),
            image_width=chunk_dict["image_width"],
            image_height=chunk_dict["image_height"],
            fps=chunk_dict["fps"],
            object_ids=chunk_dict["object_ids"],
            start_time=chunk_dict.get("start_time"),  # Backwards compatibility
        )

    @classmethod
    def from_file(cls: type[T], path: str | Path) -> T:
        """Read chunk from disk."""
        with gzip.open(path, "rt") as f:
            return cls.from_dict(json.loads(f.read()))



import numpy as np
import pytest

from sparrow_datums.boxes.frame_augmented_boxes import FrameAugmentedBoxes
from sparrow_datums.tracking.box_tracking import BoxTracking

from ..exceptions import ValidationError
from ..types import PType
from .augmented_box_tracking import AugmentedBoxTracking

DARWIN_DICT = {
    "dataset": "Ring Tracking",
    "image": {
        "width": 1920,
        "height": 1080,
        "fps": 30.0,
        "original_filename": "Ring_541700402_1820_7075017446887328011.mp4",
        "filename": "Ring_541700402_1820_7075017446887328011.mp4",
        "url": "https://darwin.v7labs.com/api/videos/4007448/original",
        "path": "/",
        "workview_url": "https://darwin.v7labs.com/workview?dataset=470324&image=1",
        "frame_count": 3,
    },
    "annotations": [
        {
            "frames": {
                "0": {
                    "bounding_box": {"h": 32.1, "w": 60.35, "x": 704.11, "y": 414.68},
                    "keyframe": True,
                },
                "1": {
                    "bounding_box": {"h": 32.1, "w": 60.35, "x": 704.11, "y": 414.68},
                    "keyframe": False,
                },
                "2": {
                    "bounding_box": {"h": 32.1, "w": 60.35, "x": 704.11, "y": 414.68},
                    "keyframe": False,
                },
            },
            "id": "f90b983d-8d31-41c3-828f-69b993099ca7",
            "interpolate_algorithm": "linear-1.1",
            "interpolated": True,
            "name": "car",
            "segments": [[0, 3783]],
        },
        {
            "frames": {
                "0": {
                    "bounding_box": {"h": 16.05, "w": 35.31, "x": 659.81, "y": 398.62},
                    "keyframe": True,
                },
                "1": {
                    "bounding_box": {"h": 16.05, "w": 35.31, "x": 659.81, "y": 398.62},
                    "keyframe": False,
                },
                "2": {
                    "bounding_box": {"h": 16.05, "w": 35.31, "x": 659.81, "y": 398.62},
                    "keyframe": False,
                },
            },
            # Leave id out of one annotation to make sure everything still works
            # "id": "c2c5e6c7-f4f5-4edb-b24a-00aa76b7674e",
            "interpolate_algorithm": "linear-1.1",
            "interpolated": True,
            "name": "car",
            "segments": [[0, 3783]],
        },
        {
            "frames": {
                "0": {
                    "bounding_box": {
                        "h": 494.39,
                        "w": 559.11,
                        "x": 987.14,
                        "y": 494.43,
                    },
                    "keyframe": True,
                },
                "1": {
                    "bounding_box": {
                        "h": 494.39,
                        "w": 559.11,
                        "x": 987.14,
                        "y": 494.43,
                    },
                    "keyframe": False,
                },
                "2": {
                    "bounding_box": {
                        "h": 494.39,
                        "w": 559.11,
                        "x": 987.14,
                        "y": 494.43,
                    },
                    "keyframe": False,
                },
            },
            "id": "8ffd95e2-0641-445a-83e5-47c664edaaa5",
            "interpolate_algorithm": "linear-1.1",
            "interpolated": True,
            "name": "bicycle",
            "segments": [[0, 3783]],
        },
    ],
}


def test_box_tracking_chunk_requires_3_dimensions():
    with pytest.raises(ValidationError):
        AugmentedBoxTracking(np.ones((2, 6)))
    with pytest.raises(ValidationError):
        AugmentedBoxTracking(np.ones((2, 2, 4)))


def test_from_dict_with_empty_data():
    boxes = AugmentedBoxTracking.from_dict(
        {
            "data": [],
            "ptype": "unknown",
            "image_width": None,
            "image_height": None,
            "fps": None,
            "object_ids": None,
        }
    )
    assert isinstance(boxes, AugmentedBoxTracking)
    assert len(boxes) == 0


def test_frame_augmented_boxes_iterator_makes_frame_augmented_boxes():
    chunk = AugmentedBoxTracking(np.ones((2, 2, 6)), ptype=PType.absolute_tlwh)
    for frame in chunk:
        assert isinstance(frame, FrameAugmentedBoxes)
        assert frame.ptype == PType.absolute_tlwh


def test_from_darwin_dict_creates_augmented_box_tracking():
    chunk = AugmentedBoxTracking.from_darwin_dict(DARWIN_DICT)
    assert chunk.shape == (3, 3, 6)
    assert isinstance(chunk, AugmentedBoxTracking)
    assert chunk.ptype == PType.absolute_tlwh
    assert chunk.image_width
    assert chunk.image_height
    np.testing.assert_equal(chunk.labels, -1)
    assert "8ffd95e2-0641-445a-83e5-47c664edaaa5" in chunk.object_ids


def test_from_frame_augmented_boxes():
    boxes_a = FrameAugmentedBoxes(np.ones((2, 6)), PType.absolute_tlwh)
    boxes_b = FrameAugmentedBoxes(np.ones((3, 6)), PType.absolute_tlwh)
    chunk = AugmentedBoxTracking.from_frame_augmented_boxes(
        [boxes_a, boxes_b], ptype=PType.absolute_tlwh
    )
    assert chunk.shape == (2, 3, 6)


def test_filter_by_class():
    chunk = AugmentedBoxTracking.from_darwin_dict(
        DARWIN_DICT, label_names=["car", "bicycle"]
    )
    cars = chunk.filter_by_class(0)
    bikes = chunk.filter_by_class(1)
    assert isinstance(cars, BoxTracking)
    assert isinstance(bikes, BoxTracking)
    assert cars.shape[1] + bikes.shape[1] == chunk.shape[1]


def test_from_box_tracking():
    chunk = AugmentedBoxTracking.from_darwin_dict(
        DARWIN_DICT, label_names=["car", "bicycle"]
    )
    cars = chunk.filter_by_class(0)
    augmented_cars = AugmentedBoxTracking.from_box_tracking(
        cars, score=1.0, class_idx=0
    )
    assert isinstance(augmented_cars, AugmentedBoxTracking)
    np.testing.assert_equal(cars.array, augmented_cars.array[..., :4])



import doctest
import os
import tempfile

import numpy as np
import pytest

from sparrow_datums.boxes.frame_boxes import FrameBoxes

from ..exceptions import ValidationError
from ..types import PType
from . import frame_augmented_boxes
from .frame_augmented_boxes import FrameAugmentedBoxes
from .single_augmented_box import SingleAugmentedBox

DARWIN_DICT = {
    "dataset": "RetinaNet Detections",
    "image": {
        "width": 1920,
        "height": 1080,
        "original_filename": "Ring_541700402_1820_7075017446887328011_00000.jpg",
        "filename": "Ring_541700402_1820_7075017446887328011_00000.jpg",
        "url": "https://darwin.v7labs.com/api/images/335989560/original",
        "thumbnail_url": "https://darwin.v7labs.com/api/images/335989560/thumbnail",
        "path": "/",
        "workview_url": "https://darwin.v7labs.com/workview?dataset=417732&image=4296",
    },
    "annotations": [
        {
            "bounding_box": {"h": 301.37, "w": 219.31, "x": 756.71, "y": 686.49},
            "name": "bicycle",
        },
        {
            "bounding_box": {"h": 480.11, "w": 527.55, "x": 1010.4, "y": 506.5},
            "name": "car",
        },
        {
            "bounding_box": {"h": 104.65, "w": 80.25, "x": 243.55, "y": 592.7},
            "name": "chair",
        },
        {
            "bounding_box": {"h": 30.25, "w": 59.21, "x": 704.58, "y": 418.17},
            "name": "car",
        },
        {
            "bounding_box": {"h": 46.23, "w": 74.03, "x": 1796.27, "y": 482.56},
            "name": "car",
        },
        {
            "bounding_box": {"h": 57.3, "w": 89.47, "x": 129.04, "y": 582.63},
            "name": "bicycle",
        },
        {
            "bounding_box": {"h": 16.48, "w": 34.51, "x": 658.83, "y": 397.2},
            "name": "car",
        },
    ],
}


def test_docstring_example():
    result = doctest.testmod(frame_augmented_boxes)
    assert result.attempted > 0
    assert result.failed == 0


def test_frame_augmented_boxes_conversion_creates_frame_augmented_boxes():
    boxes_a = FrameAugmentedBoxes(np.ones((2, 6)), ptype=PType.absolute_tlwh)
    boxes_b = boxes_a.to_tlbr()
    assert boxes_b.is_tlbr
    assert isinstance(boxes_b, FrameAugmentedBoxes)


def test_frame_augmented_boxes_with_bad_shape_throws_value_error():
    with pytest.raises(ValidationError):
        FrameAugmentedBoxes(np.ones(6))
    with pytest.raises(ValidationError):
        FrameAugmentedBoxes(np.ones((2, 4)))


def test_frame_augmented_boxes_iterator_makes_frame_augmented_boxes():
    boxes = FrameAugmentedBoxes(np.ones((2, 6)), ptype=PType.absolute_tlwh)
    for box in boxes:
        assert isinstance(box, SingleAugmentedBox)
        assert box.ptype == PType.absolute_tlwh


def test_frame_augmented_boxes_can_become_darwin_dict():
    boxes = FrameAugmentedBoxes(np.ones((2, 6)), ptype=PType.absolute_tlwh)
    data = boxes.to_darwin_dict("foobar")
    assert data["image"]["filename"] == "foobar"
    for annotation in data["annotations"]:
        assert "bounding_box" in annotation
        assert annotation["name"] == "Unknown"


def test_darwin_dict_is_json_serializable():
    boxes = FrameAugmentedBoxes(np.ones((2, 6)), ptype=PType.absolute_tlwh)
    with tempfile.TemporaryDirectory() as tmpdir:
        path = os.path.join(tmpdir, "test.json")
        boxes.to_darwin_file(path, "foobar")


def test_from_darwin_dict_creates_frame_augmented_boxes():
    boxes = FrameAugmentedBoxes.from_darwin_dict(DARWIN_DICT)
    assert len(boxes) == 7
    assert isinstance(boxes, FrameAugmentedBoxes)
    assert boxes.ptype == PType.absolute_tlwh
    assert boxes.image_width
    assert boxes.image_height
    np.testing.assert_equal(boxes.labels, -1)


def test_from_darwin_dict_creates_frame_augmented_boxes_with_no_boxes():
    empty_darwin_dict = DARWIN_DICT.copy()
    empty_darwin_dict["annotations"] = []
    boxes = FrameAugmentedBoxes.from_darwin_dict(empty_darwin_dict)
    assert isinstance(boxes, FrameAugmentedBoxes)


def test_from_dict_with_empty_data():
    boxes = FrameAugmentedBoxes.from_dict(
        {
            "data": [],
            "ptype": "unknown",
            "image_width": None,
            "image_height": None,
            "fps": None,
            "object_ids": None,
        }
    )
    assert isinstance(boxes, FrameAugmentedBoxes)
    assert len(boxes) == 0


def test_from_darwin_dict_creates_frame_augmented_boxes_with_non_bounding_box_annotations():
    non_box_dict = DARWIN_DICT.copy()
    non_box_dict["annotations"] = [{"foo": "bar"}]
    boxes = FrameAugmentedBoxes.from_darwin_dict(non_box_dict)
    assert isinstance(boxes, FrameAugmentedBoxes)


def test_from_single_box():
    box = SingleAugmentedBox(np.ones(6), PType.absolute_tlbr)
    frame_augmented_boxes = FrameAugmentedBoxes.from_single_box(box)
    assert frame_augmented_boxes.shape == (1, 6)
    assert frame_augmented_boxes.ptype == PType.absolute_tlbr


def test_add_box():
    box = SingleAugmentedBox(np.ones(6), PType.absolute_tlbr)
    frame_augmented_boxes = FrameAugmentedBoxes.from_single_box(box)
    frame_augmented_boxes = frame_augmented_boxes.add_box(box)
    assert frame_augmented_boxes.shape == (2, 6)


def test_get_single_box():
    labels = ["car", "bicycle", "chair"]
    boxes = FrameAugmentedBoxes.from_darwin_dict(DARWIN_DICT, label_names=labels)
    box = boxes.get_single_box(-1)
    assert box.shape == (6,)
    assert labels[box.label] == "car"


def test_to_frame_boxes():
    boxes = FrameAugmentedBoxes.from_darwin_dict(DARWIN_DICT)
    boxes = boxes.to_frame_boxes()
    assert isinstance(boxes, FrameBoxes)
    assert boxes.ptype == PType.absolute_tlwh
    assert boxes.shape[-1] == 4



import numpy as np
import pytest

from ..exceptions import ValidationError
from ..types import PType
from .single_box import SingleBox


def test_single_box_conversion_creates_single_box():
    box_a = SingleBox(np.ones(4), PType.absolute_tlbr)
    box_b = box_a.to_tlwh()
    assert box_b.is_tlwh
    assert isinstance(box_b, SingleBox)


def test_single_box_with_bad_shape_throws_value_error():
    with pytest.raises(ValidationError):
        SingleBox(np.ones((2, 4)))
    with pytest.raises(ValidationError):
        SingleBox(np.ones(3))



"""Data types."""
import enum

import numpy as np
import numpy.typing as npt

FloatArray = npt.NDArray[np.float64]


class PType(enum.Enum):
    """Enum for chunk parameterization types."""

    unknown = "unknown"
    """
    Unknown parameterization
    """

    absolute_tlbr = "absolute_tlbr"
    """
    (x1, y1, x2, y2) in absolute pixel coordinates
    """

    absolute_tlwh = "absolute_tlwh"
    """
    (x, y, w, h) in absolute pixel coordinates
    """

    relative_tlbr = "relative_tlbr"
    """
    (x1, y1, x2, y2) in relative pixel coordinates
    """

    relative_tlwh = "relative_tlwh"
    """
    (x, y, w, h) in relative pixel coordinates
    """

    @property
    def is_relative(self) -> bool:
        """Whether the parameterization is in relative space."""
        return "relative" in self.name

    @property
    def is_absolute(self) -> bool:
        """Whether the parameterization is in absolute space."""
        return "absolute" in self.name

    @property
    def is_tlbr(self) -> bool:
        """Whether the parameterization is TLBR."""
        return "tlbr" in self.name

    @property
    def is_tlwh(self) -> bool:
        """Whether the parameterization is TLWH."""
        return "tlwh" in self.name

    @property
    def as_relative(self) -> "PType":
        """Convert parameterization type to relative."""
        return PType(self.name.replace("absolute", "relative"))

    @property
    def as_absolute(self) -> "PType":
        """Convert parameterization type to absolute."""
        return PType(self.name.replace("relative", "absolute"))

    @property
    def as_tlbr(self) -> "PType":
        """Convert parameterization type to TLBR."""
        return PType(self.name.replace("tlwh", "tlbr"))

    @property
    def as_tlwh(self) -> "PType":
        """Convert parameterization type to TLWH."""
        return PType(self.name.replace("tlbr", "tlwh"))



import json

from sparrow_datums import FrameAugmentedBoxes, PType

from .types import ChunkPath, Header


def test_header_serialization():
    value_str = Header(
        classname=FrameAugmentedBoxes.__name__,
        ptype=PType.absolute_tlbr.name,
        image_width=None,
        image_height=None,
        fps=1,
        start_time=0,
    ).to_json()
    value = json.loads(value_str)
    assert value["classname"] == "FrameAugmentedBoxes"
    assert value["image_height"] is None


def test_chunk_path_serialization():
    value_dict = {"path": "0000.json.gz", "duration": 4.0, "start_time": 0}
    value = ChunkPath.from_dict(value_dict)
    assert value.duration == 4.0



import numpy as np

from ..types import PType
from .frame_augmented_boxes import FrameAugmentedBoxes
from .nms import non_max_suppression


def test_nms_filters_by_score():
    data = np.array([[0, 0, 1, 1, 0.4, 0], [1, 1, 1, 1, 0.6, 1]])
    boxes = FrameAugmentedBoxes(data, ptype=PType.absolute_tlwh)
    filtered_boxes = non_max_suppression(boxes)
    assert len(filtered_boxes) == 1


def test_nms_removes_overlapping_boxes_by_class():
    data = np.array(
        [
            [0, 0, 1, 1, 0.4, 0],
            [0.1, 0.1, 1, 1, 0.65, 0],
            [1, 1, 1, 1, 0.7, 1],
            [1.1, 1.1, 1, 1, 0.6, 1],
        ]
    )
    boxes = FrameAugmentedBoxes(data, ptype=PType.absolute_tlwh)
    filtered_boxes = non_max_suppression(boxes)
    assert set(filtered_boxes.scores) == {0.65, 0.7}


def test_nms_handles_empty_return():
    data = np.array([[0, 0, 1, 1, 0.4, 0]])
    boxes = FrameAugmentedBoxes(data, ptype=PType.absolute_tlwh)
    filtered_boxes = non_max_suppression(boxes)
    assert len(filtered_boxes) == 0


def test_nms_handles_empty_input():
    data = np.zeros((0, 6))
    boxes = FrameAugmentedBoxes(data, ptype=PType.absolute_tlwh)
    filtered_boxes = non_max_suppression(boxes)
    assert len(filtered_boxes) == 0
